#define BUILD_NUMBER \
	"build-2822639"
#define BUILD_NUMBER_NUMERIC \
	2822639
#define BUILD_NUMBER_NUMERIC_STRING \
	"2822639"
#define PRODUCT_BUILD_NUMBER \
	"product-build-48224"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	48224
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"48224"
